/* WaterEngine.cpp
Implements the Water Engine
(C) 2000 N. Douglas
12th April 2000
*/

#include "WaterEngine.h"
#include "Preferences.h"
#include <math.h>
#include <assert.h>

WaterEngine::WaterEngine(u32 _mols, Bitmap *bmp, Preferences *_prefs)
{
	scrimage=bmp;
	mols=0;
	densities=false;
	mol=NULL;
	prefs=_prefs;
	if(prefs->normalmolcol==0)
		prefs->normalmolcol=0x00ff0000;
	if(prefs->lightmolcol==0)
		prefs->lightmolcol=0x000000ff;
	
	setNoOfMols(_mols, 2);
}

WaterEngine::~WaterEngine()
{
	if(mol)
	{
		free(mol);
		mol=NULL;
	}
	mols=0;
	scrimage=NULL;
}

// Here follow some highly optimised code (most intensely called in Flow)
#pragma check_stack(off)
#pragma inline_depth(4)
#pragma optimize("py", off)
#pragma optimize("agtw", on)
ENGOPTS bool ENGOPTS2 WaterEngine::_testdown(u32 *curoffset, const u32 width, const u32 height, u32 oldscrcol, bool light) const
{
	u32 b;
	/* If pixel below empty, move down */
	b=scrimage->readpixel(*curoffset+width);
	if(b==0 || b==pumpLscrcol || b==pumpRscrcol || (filter && b==filtAscrcol)
		|| (light && b==filtBscrcol) || (!light && b==filtRscrcol))
	{
		*curoffset+=width;
		if(*curoffset>((height-1)*width)) *curoffset-=(height-2)*width;
		return true;
	}
	return false;
}

ENGOPTS bool ENGOPTS2 WaterEngine::_testlr(u32 *curoffset, const u32 width, const u32 height, u32 oldscrcol, bool light)
{
	u32 l,r;

	l=scrimage->readpixel(*curoffset-1);
	r=scrimage->readpixel(*curoffset+1);
	if(l==pumpLscrcol || l==pumpUscrcol || (filter && l==filtAscrcol)
		|| (light && l==filtBscrcol) || (!light && l==filtRscrcol))
		l=0;
	if(r==pumpRscrcol || r==pumpUscrcol || (filter && r==filtAscrcol)
		|| (light && r==filtBscrcol) || (!light && r==filtRscrcol))
		r=0;
	/* If left and right empty, move random */
	if(l==0 && r==0)
	{
		static unsigned int seed=0x12345678;
		int t;
		t=(seed & (1<<31))!=0;
		seed<<=1;
		if(t)
		{
			seed^=0x1d872b41;
			*curoffset+=1;
		}
		else
		{
			*curoffset-=1;
		}
		return true;
	}
	/* If left free, move left. If right free, move right */
	if(l==0	&& oldscrcol!=pumpRscrcol)
	{
		*curoffset-=1;
		return true;
	}
	if(r==0 && oldscrcol!=pumpLscrcol)
	{
		*curoffset+=1;
		return true;
	}
	return false;
}

ENGOPTS bool ENGOPTS2 WaterEngine::_testup(u32 *curoffset, const u32 width, const u32 height, u32 oldscrcol, bool light) const
{
	if(*curoffset>=width)
	{
		u32 abovecol;
		abovecol=scrimage->readpixel(*curoffset-width);
		if(abovecol==0 || abovecol==pumpUscrcol || abovecol==pumpLscrcol || abovecol==pumpRscrcol
			|| (filter && abovecol==filtAscrcol) || (light && abovecol==filtBscrcol) || (!light && abovecol==filtRscrcol))
		{
			*curoffset-=width;
			return true;
		}
		/* If I am a "light" molecule, make me rise if possible */
		if(densities && light)
		{
			s32 peekoffset;
			u32 col=1, mycol;
			bool onlylight=true;
			mycol=scrimage->readpixel(*curoffset);
			peekoffset=*curoffset-width-width;
			while(peekoffset>=(s32) width)
			{
				col=scrimage->readpixel(peekoffset);
				if(col!=0 && col!=mycol) onlylight=false;
				if(col!=0 && col!=255)
					peekoffset-=width;
				else
					break;
			}
			if(!onlylight && col==0)
			{
				*curoffset=peekoffset;
				return true;
			}
		}
	}
	return false;
}

void WaterEngine::update()
{
	u32 m;


#if 0
	{
		int c=0;
		for(int y=0; y<8; y++)
		{
			for(int x=0; x<32; x++)
			{
				int off;
				off=(scrimage->width*(50+y*2))+300+x*2;
				scrimage->writepixel(off, c);
				scrimage->writepixel(off+1, c);
				scrimage->writepixel(off+scrimage->width, c);
				scrimage->writepixel(off+scrimage->width+1, c);
				c++;
			}
		}
	}
#endif

	for(m=0; m<mols; m++)
	{
		Molecule *w;
		u32 curoffset, oldoffset;
		u32 oldscrcol;
		u32 width, height;
		bool light;
		w=&mol[m];
		light=w->light;
		curoffset=oldoffset=w->pixeloffset;
		oldscrcol=w->oldscrcol;
		width=scrimage->width;
		height=scrimage->height;

		do
		{
			if(oldscrcol==pumpUscrcol)
			{
				if(_testup(&curoffset, width, height, oldscrcol, light)) break;
				if(_testlr(&curoffset, width, height, oldscrcol, light)) break;
				break;
			}
			if(oldscrcol==pumpLscrcol || oldscrcol==pumpRscrcol)
			{
				if(_testlr(&curoffset, width, height, oldscrcol, light)) break;
				if(_testup(&curoffset, width, height, oldscrcol, light)) break;
				if(_testdown(&curoffset, width, height, oldscrcol, light)) break;
				break;
			}
			/* If pixel below empty, move down */
			if(_testdown(&curoffset, width, height, oldscrcol, light)) break;
			/* Can I move left or right? */
			if(_testlr(&curoffset, width, height, oldscrcol, light)) break;
			/* If left and right full, can I move up? */
			if(_testup(&curoffset, width, height, oldscrcol, light)) break;
			break;

		} while(0);

#ifdef _DEBUG
		assert(curoffset>=0 && curoffset<width*height);
#endif
		scrimage->writepixel(oldoffset, oldscrcol);
		w->oldscrcol=scrimage->readpixel(curoffset);
		scrimage->writepixel(curoffset, w->scrcol);
		w->pixeloffset=curoffset;
	}
}
#pragma optimize("", on)
#pragma check_stack
// End of highly optimised code

void WaterEngine::render(bool plot)
{
	u32 m;

	for(m=0; m<mols; m++)
	{
		Molecule *w;
		w=&mol[m];
		if(plot)
		{
			w->oldscrcol=scrimage->readpixel(w->pixeloffset);
			scrimage->writepixel(w->pixeloffset, w->scrcol);
		}
		else
			scrimage->writepixel(w->pixeloffset, w->oldscrcol);
	}
}

void WaterEngine::palettechanged()
{
	// Run through the molecules, calculating RGB->physical colours
	// ->RGBtoPhysical() is slow, so we hold the last two colours translated
	u32 m, oldrgb[2]={0,0}, oldphys[2]={0,0};


	for(m=0; m<mols; m++)
	{
		if(mol[m].pixelcolour==oldrgb[0])
			mol[m].scrcol=oldphys[0];
		else if(mol[m].pixelcolour==oldrgb[1])
			mol[m].scrcol=oldphys[1];
		else
		{
			oldrgb[1]=oldrgb[0];
			oldphys[1]=oldphys[0];
			mol[m].scrcol=oldphys[0]=scrimage->RGBtoPhysical((oldrgb[0]=mol[m].pixelcolour));
		}
	}
	// Also determine what the magic colours are
	pumpUscrcol=scrimage->RGBtoPhysical(0x00808000);
	pumpLscrcol=scrimage->RGBtoPhysical(0x00008000);
	pumpRscrcol=scrimage->RGBtoPhysical(0x00000080);
	filtAscrcol=scrimage->RGBtoPhysical(0x00c0c0c0);	// light grey
	filtRscrcol=scrimage->RGBtoPhysical(0x00c0c0f0);	// white-red (251)
	filtBscrcol=scrimage->RGBtoPhysical(0x00f0c0c0);	// white-blue (219)
}

void WaterEngine::setSize(u32 w, u32 h)
{
	if(w<scrimage->width || h<scrimage->height)
	{
		DECLARE_ERROR(e, "Not implemented yet");
		throw e;
	}
	// Adjust all the pixel offsets
	for(u32 m=0; m<mols; m++)
	{
		u32 x,y;
		y=mol[m].pixeloffset / scrimage->width;
		x=mol[m].pixeloffset - (y*scrimage->width);
		mol[m].pixeloffset=y*w+x;
	}
	// Extend the bitmap
	scrimage->setSize(w, h);
}

void WaterEngine::setNoOfMols(u32 newmols, int col)
{
	u32 m, startoffset;
	Molecule *newmol=NULL;

	if(newmols)
	{
		if(!(newmol=(Molecule *) calloc(newmols+2, sizeof(Molecule))))
		{
			// Not enough memory for requested molecules
			DECLARE_ERROR(e, "Not enough memory for requested molecule number");
			throw e;
		}
		startoffset=((scrimage->width*1)+1);
		m=0;
		do
		{
			if(m<mols)
			{
				// Copy existing molecule
				newmol[m]=mol[m];
			}
			else
			{
				// Create a new molecule at the top left
				newmol[m].pixeloffset=startoffset;
				if(col==0)
					newmol[m].light=false;
				else if(col==1)
					newmol[m].light=true;
				else
				{
					if(m & 1)
						newmol[m].light=true;
					else
						newmol[m].light=false;
				}
				if(newmol[m].light)
					newmol[m].pixelcolour=prefs->lightmolcol;
				else
					newmol[m].pixelcolour=prefs->normalmolcol;
				newmol[m].scrcol=0xffffffff;
				newmol[m].oldscrcol=0;
			}
		} while(m++<newmols);
	}
	if(mol)
	{
		scrimage->getAddr();
		render(0);
		scrimage->releaseAddr();
		free(mol);
		mol=NULL;
		mols=0;
	}
	mol=newmol;
	mols=newmols;

	// Determine best fit colours for the molecules
	palettechanged();
}

bool WaterEngine::load(HANDLE inh)
{
	u32 t;
	DWORD r;
	// Check for validity
	TEST_WIN32ERR(!ReadFile(inh, &t, sizeof(u32), &r, NULL));
	if(t!=*(u32 *) "Flow")
	{
		DECLARE_ERROR(e, "This file is not a valid Flow game file");
		throw e;
	}
	
	// Remove molecules
	setNoOfMols(0, 2);

	// Load the bitmap
	scrimage->load(inh);

	// Now load all the molecules
	TEST_WIN32ERR(!ReadFile(inh, &t, sizeof(u32), &r, NULL));
	if(t!=*(u32 *) "MOLS") return true;
	TEST_WIN32ERR(!ReadFile(inh, &t, sizeof(u32), &r, NULL));
	if(t==0)
		setNoOfMols(100, 2);
	else
		setNoOfMols(t, 2);
	for(u32 m=0; m<mols; m++)
	{
		TEST_WIN32ERR(!ReadFile(inh, &mol[m].pixeloffset, sizeof(u32), &r, NULL));
		TEST_WIN32ERR(!ReadFile(inh, &mol[m].pixelcolour, sizeof(u32), &r, NULL));
		mol[m].scrcol=mol[m].oldscrcol=0;
	}
	// Determine best fit colours for the molecules & plot
	palettechanged();
	scrimage->getAddr();
	render(1);
	scrimage->releaseAddr();
	return t==0;
}

void WaterEngine::save(HANDLE outh, bool autoset)
{
#ifdef SHAREWARE
	DECLARE_ERROR(e, "Not implemented in the shareware build");
	throw e;
#else
	DWORD w;
	TEST_WIN32ERR(!WriteFile(outh, "Flow", 4, &w, NULL));
	
	// Save the bitmap
	scrimage->getAddr();
	render(0);
	scrimage->save(outh);
	render(1);
	scrimage->releaseAddr();

	// Now save all the molecule positions
	if(!autoset)
	{
		TEST_WIN32ERR(!WriteFile(outh, "MOLS", 4, &w, NULL));
		TEST_WIN32ERR(!WriteFile(outh, &mols, sizeof(u32), &w, NULL));
		for(u32 m=0; m<mols; m++)
		{
			TEST_WIN32ERR(!WriteFile(outh, &mol[m].pixeloffset, sizeof(u32), &w, NULL));
			TEST_WIN32ERR(!WriteFile(outh, &mol[m].pixelcolour, sizeof(u32), &w, NULL));
		}
	}
#endif
}
